# pytest . -v -s --log-cli-level=INFO

import asyncio
import logging
import pytest
import pytest_asyncio
from dotenv import dotenv_values

from pymexc.proto import (
    PrivateAccountV3Api,
    PrivateDealsV3Api,
    PrivateOrdersV3Api,
    PublicAggreBookTickerV3Api,
    PublicAggreDealsV3Api,
    PublicAggreDepthsV3Api,
    PublicLimitDepthsV3Api,
    PublicMiniTickerV3Api,
    PublicMiniTickersV3Api,
    PublicSpotKlineV3Api,
    PublicBookTickerBatchV3Api,
)
from pymexc._async.spot import WebSocket, HTTP
from pymexc._async.base_websocket import SPOT as SPOT_WS

env = dotenv_values(".env")
api_key = env["API_KEY"]
api_secret = env["API_SECRET"]

PRINT_RESPONSE = True
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s:%(lineno)d - %(levelname)s - %(message)s",  # add line number
)
logger = logging.getLogger(__name__)

listen_key = None


@pytest_asyncio.fixture
async def ws_client():
    """Create a new WebSocket client for each test."""
    global listen_key
    if not listen_key:
        listen_key = await HTTP(api_key=api_key, api_secret=api_secret).create_listen_key()
        listen_key = listen_key.get("listenKey")

    client = WebSocket(api_key=api_key, api_secret=api_secret, proto=True, listenKey=listen_key)
    logger.info(f"WebSocket client created: {client}")
    yield client
    # Cleanup
    if hasattr(client, "ws"):
        logger.info(f"Closing WebSocket: {client.ws}")
        await client.ws.close()
    if hasattr(client, "session"):
        logger.info(f"Closing HTTP session: {client.session}")
        await client.session.close()


async def make_mxusdt_order(time_wait: int = 2):
    """Make a MXUSDT order and cancel it."""
    client = HTTP(api_key=api_key, api_secret=api_secret)
    await asyncio.sleep(time_wait)
    order_response = await client.order("MXUSDT", "BUY", "LIMIT", quantity=20, price=0.1)
    logger.info(f"MXUSDT Order Response: {order_response}")
    await asyncio.sleep(time_wait)
    cancel_order_response = await client.cancel_order("MXUSDT", order_response["orderId"])
    logger.info(f"MXUSDT Cancel Order Response: {cancel_order_response}")
    await asyncio.sleep(time_wait)


@pytest.mark.asyncio
async def test_deals_stream(ws_client: WebSocket):
    """Test the deals stream subscription."""
    messages: list[PublicAggreDealsV3Api] = []

    async def handle_message(msg):
        messages.append(msg)
        logger.info(f"Received message: {msg}")
        if len(messages) >= 2:  # Get at least 2 messages
            await ws_client.unsubscribe(ws_client.deals_stream)

    await ws_client.deals_stream(handle_message, "BTCUSDT", interval="100ms")

    # Wait for messages
    await asyncio.sleep(5)

    if PRINT_RESPONSE:
        logger.info(f"{messages[0].deals[0].price=}")

    assert len(messages) >= 2
    for msg in messages:
        assert isinstance(msg, PublicAggreDealsV3Api)
        assert float(msg.deals[0].price) > 0
        assert float(msg.deals[0].quantity) > 0
        assert msg.deals[0].tradeType > 0
        assert int(msg.deals[0].time) > 0


@pytest.mark.asyncio
async def test_kline_stream(ws_client: WebSocket):
    """Test the kline stream subscription."""
    messages: list[PublicSpotKlineV3Api] = []

    async def handle_message(msg: PublicSpotKlineV3Api):
        messages.append(msg)
        logger.info(f"Received message: {msg}")
        if len(messages) >= 2:  # Get at least 2 messages
            await ws_client.unsubscribe("public.kline")

    await ws_client.kline_stream(handle_message, "BTCUSDT", interval="Min1")

    # Wait for messages
    await asyncio.sleep(2)

    assert len(messages) >= 2
    for msg in messages:
        assert isinstance(msg, PublicSpotKlineV3Api)
        assert msg.interval == "Min1"
        assert int(msg.windowStart) > 0
        assert float(msg.openingPrice) > 0
        assert float(msg.closingPrice) > 0
        assert float(msg.highestPrice) > 0
        assert float(msg.lowestPrice) > 0
        assert float(msg.volume) > 0
        assert float(msg.amount) > 0
        assert int(msg.windowEnd) > 0


@pytest.mark.asyncio
async def test_increase_depth_stream(ws_client: WebSocket):
    """Test the increase depth stream subscription."""
    messages: list[PublicAggreDepthsV3Api] = []

    async def handle_message(msg: PublicAggreDepthsV3Api):
        messages.append(msg)
        # logger.info(f"Received message: {msg}")
        if len(messages) >= 2:  # Get at least 2 messages
            await ws_client.unsubscribe(ws_client.depth_stream)

    await ws_client.depth_stream(handle_message, "BTCUSDT")

    # Wait for messages
    await asyncio.sleep(2)

    assert len(messages) >= 2
    for msg in messages:
        assert isinstance(msg, PublicAggreDepthsV3Api)
        assert msg.eventType == "spot@public.aggre.depth.v3.api.pb@100ms"
        assert msg.fromVersion is not None
        assert len(msg.asks) > 0
        assert len(msg.bids) > 0


@pytest.mark.asyncio
async def test_limit_depth_stream(ws_client: WebSocket):
    """Test the limit depth stream subscription."""
    messages: list[PublicLimitDepthsV3Api] = []

    async def handle_message(msg: PublicLimitDepthsV3Api):
        messages.append(msg)
        logger.info(f"Received message: {msg}")
        if len(messages) >= 2:  # Get at least 2 messages
            await ws_client.unsubscribe(ws_client.limit_depth_stream)

    await ws_client.limit_depth_stream(handle_message, "BTCUSDT", 5)

    # Wait for messages
    await asyncio.sleep(3)

    assert len(messages) >= 2
    for msg in messages:
        assert isinstance(msg, PublicLimitDepthsV3Api)
        assert msg.eventType == "spot@public.limit.depth.v3.api.pb"
        assert msg.version is not None
        assert len(msg.asks) > 0
        assert len(msg.bids) > 0


@pytest.mark.asyncio
async def test_book_ticker_stream(ws_client: WebSocket):
    """Test the book ticker stream subscription."""
    messages: list[PublicAggreBookTickerV3Api] = []

    async def handle_message(msg: PublicAggreBookTickerV3Api):
        messages.append(msg)
        logger.info(f"Received message: {msg}")
        if len(messages) >= 2:  # Get at least 2 messages
            await ws_client.unsubscribe(ws_client.book_ticker_stream)

    await ws_client.book_ticker_stream(handle_message, "BTCUSDT")

    # Wait for messages
    await asyncio.sleep(2)

    assert len(messages) >= 2
    for msg in messages:
        assert isinstance(msg, PublicAggreBookTickerV3Api)
        assert msg.bidPrice is not None
        assert float(msg.bidPrice) > 0
        assert msg.bidQuantity is not None
        assert float(msg.bidQuantity) > 0
        assert msg.askPrice is not None
        assert float(msg.askPrice) > 0
        assert msg.askQuantity is not None
        assert float(msg.askQuantity) > 0


@pytest.mark.asyncio
async def test_book_ticker_batch_stream(ws_client: WebSocket):
    """Test the book ticker batch stream subscription."""
    messages: list[PublicBookTickerBatchV3Api] = []

    async def handle_message(msg: PublicBookTickerBatchV3Api):
        messages.append(msg)
        logger.info(f"Received message: {msg}")
        if len(messages) >= 2:  # Get at least 2 messages
            await ws_client.unsubscribe(ws_client.book_ticker_batch_stream)

    await ws_client.book_ticker_batch_stream(handle_message, "BTCUSDT")

    # Wait for messages
    await asyncio.sleep(2)

    assert len(messages) >= 2
    for msg in messages:
        assert isinstance(msg, PublicBookTickerBatchV3Api)
        assert msg.items[0].bidPrice is not None
        assert float(msg.items[0].bidPrice) > 0
        assert msg.items[0].bidQuantity is not None
        assert float(msg.items[0].bidQuantity) > 0
        assert msg.items[0].askPrice is not None
        assert float(msg.items[0].askPrice) > 0
        assert msg.items[0].askQuantity is not None
        assert float(msg.items[0].askQuantity) > 0


@pytest.mark.asyncio
async def test_mini_tickers_stream(ws_client: WebSocket):
    """Test the mini tickers stream subscription."""
    messages: list[PublicMiniTickersV3Api] = []
    done = asyncio.Event()

    async def handle_message(msg: PublicMiniTickersV3Api):
        messages.append(msg)
        logger.info(f"Received mini tickers message: {str(msg)[:50]}")
        if len(messages) >= 1:  # Mini tickers push every 3 seconds, so wait for at least 1
            await ws_client.unsubscribe(ws_client.mini_tickers_stream)
            done.set()

    await ws_client.mini_tickers_stream(handle_message, timezone="UTC+8")

    # Wait for messages
    try:
        await asyncio.wait_for(done.wait(), timeout=15)  # Increased timeout for 3-second interval
    except asyncio.TimeoutError:
        pytest.fail("Did not receive mini tickers messages")

    assert len(messages) >= 1
    for msg in messages:
        assert isinstance(msg, PublicMiniTickersV3Api)
        assert hasattr(msg, "items") and msg.items
        assert len(msg.items) > 0
        first = msg.items[0]
        assert hasattr(first, "symbol") and first.symbol
        assert hasattr(first, "price") and first.price is not None


@pytest.mark.asyncio
async def test_mini_ticker_stream(ws_client: WebSocket):
    """Test the mini ticker stream subscription."""
    messages: list[PublicMiniTickerV3Api] = []
    done = asyncio.Event()

    async def handle_message(msg: PublicMiniTickerV3Api):
        messages.append(msg)
        logger.info(f"Received mini ticker message: {str(msg)[:50]}")
        if len(messages) >= 1:  # Mini ticker pushes every 3 seconds, so wait for at least 1
            await ws_client.unsubscribe(ws_client.mini_ticker_stream)
            done.set()

    await ws_client.mini_ticker_stream(handle_message, "BTCUSDT", timezone="UTC+8")

    # Wait for messages
    try:
        await asyncio.wait_for(done.wait(), timeout=15)  # Increased timeout for 3-second interval
    except asyncio.TimeoutError:
        pytest.fail("Did not receive mini ticker messages")

    assert len(messages) >= 1
    for msg in messages:
        assert isinstance(msg, PublicMiniTickerV3Api)
        assert hasattr(msg, "symbol") and msg.symbol == "BTCUSDT"
        assert hasattr(msg, "price") and msg.price is not None


@pytest.mark.asyncio
async def test_account_update(ws_client: WebSocket):
    """Test the account update stream subscription."""
    messages: list[PrivateAccountV3Api] = []

    async def handle_message(msg: PrivateAccountV3Api):
        messages.append(msg)
        logger.info(f"Received message: {msg}")
        if len(messages) >= 1:  # Get at least 1 message
            await ws_client.unsubscribe("private.account")

    await ws_client.account_update(handle_message)

    await make_mxusdt_order()

    # Wait for messages
    await asyncio.sleep(2)

    assert len(messages) >= 1
    for msg in messages:
        assert isinstance(msg, PrivateAccountV3Api)
        assert msg.vcoinName == "USDT"
        assert msg.coinId == "128f589271cb4951b03e71e6323eb7be"
        assert msg.balanceAmount is not None
        assert msg.balanceAmountChange is not None
        assert msg.frozenAmount is not None
        assert msg.frozenAmountChange is not None
        assert msg.type == "ENTRUST_PLACE"
        assert msg.time > 0


@pytest.mark.asyncio
async def test_account_deals(ws_client: WebSocket):
    """Test the account deals stream subscription."""
    messages: list[PrivateDealsV3Api] = []

    async def handle_message(msg):
        messages.append(msg)
        logger.info(f"Received message: {msg}")
        if len(messages) >= 2:  # Get at least 2 messages
            await ws_client.unsubscribe(ws_client.account_deals)

    await ws_client.account_deals(handle_message)

    await make_mxusdt_order()

    # Wait for messages
    await asyncio.sleep(2)

    if PRINT_RESPONSE:
        logger.info(f"{messages}")

    await ws_client.unsubscribe(ws_client.account_deals)

    await asyncio.sleep(2)

    assert len(messages) >= 0
    for msg in messages:
        assert isinstance(msg, PrivateDealsV3Api)
        assert float(msg.price) > 0
        assert float(msg.quantity) > 0
        assert msg.tradeType > 0
        assert int(msg.time) > 0


@pytest.mark.asyncio
async def test_account_orders(ws_client: WebSocket):
    """Test the account orders stream subscription."""
    messages: list[PrivateOrdersV3Api] = []

    async def handle_message(msg: PrivateOrdersV3Api):
        messages.append(msg)
        logger.info(f"Received message: {msg}")
        if len(messages) >= 1:  # Get at least 1 message
            await ws_client.unsubscribe("private.orders")

    await ws_client.account_orders(handle_message)

    await make_mxusdt_order()

    # Wait for messages
    await asyncio.sleep(2)

    assert len(messages) >= 1
    for msg in messages:
        assert isinstance(msg, PrivateOrdersV3Api)
        assert msg.id is not None
        assert msg.clientId is not None
        assert msg.price is not None
        assert msg.quantity is not None
        assert msg.amount is not None
        assert msg.avgPrice is not None
        assert msg.orderType is not None
        assert msg.tradeType is not None
        assert msg.isMaker is not None
        assert msg.remainAmount is not None
        assert msg.remainQuantity is not None
        assert msg.lastDealQuantity is not None
        assert msg.cumulativeQuantity is not None
        assert msg.cumulativeAmount is not None
        assert msg.status is not None
        assert msg.createTime is not None
        assert msg.market is not None
        assert msg.triggerType is not None
        assert msg.triggerPrice is not None
        assert msg.state is not None
        assert msg.ocoId is not None
        assert msg.routeFactor is not None
        assert msg.symbolId is not None
        assert msg.marketId is not None
        assert msg.marketCurrencyId is not None
        assert msg.currencyId is not None


@pytest.mark.asyncio
async def test_multiple_subscriptions(ws_client: WebSocket):
    """Test multiple stream subscriptions."""
    messages: list[PublicAggreDealsV3Api | PublicAggreBookTickerV3Api] = []
    done = asyncio.Event()

    async def handle_message(msg: PublicAggreDealsV3Api | PublicAggreBookTickerV3Api):
        messages.append(msg)
        logger.info(f"Received message type: {type(msg)}")
        if len(messages) >= 20:  # Get at least 10 messages from each stream
            await ws_client.unsubscribe(ws_client.deals_stream, ws_client.book_ticker_stream)
            done.set()

    # Subscribe to multiple streams
    await ws_client.deals_stream(handle_message, "BTCUSDT")
    await ws_client.book_ticker_stream(handle_message, "BTCUSDT")

    # Wait for messages
    try:
        await asyncio.wait_for(done.wait(), timeout=15)
    except asyncio.TimeoutError:
        pytest.fail("Did not receive messages from both streams")

    assert len(messages) >= 4
    # Verify we got both types of messages
    message_types = {type(msg) for msg in messages}

    assert PublicAggreDealsV3Api in message_types
    assert PublicAggreBookTickerV3Api in message_types

    for msg in messages:
        if isinstance(msg, PublicAggreDealsV3Api):
            assert msg.deals[0].price is not None
            assert float(msg.deals[0].price) > 0
            assert msg.deals[0].quantity is not None
            assert float(msg.deals[0].quantity) > 0
            assert msg.deals[0].tradeType is not None
            assert msg.deals[0].tradeType > 0
            assert msg.deals[0].time is not None
            assert int(msg.deals[0].time) > 0

        if isinstance(msg, PublicAggreBookTickerV3Api):
            assert msg.bidPrice is not None
            assert float(msg.bidPrice) > 0
            assert msg.bidQuantity is not None
            assert float(msg.bidQuantity) > 0
            assert msg.askPrice is not None
            assert float(msg.askPrice) > 0
            assert msg.askQuantity is not None
            assert float(msg.askQuantity) > 0



import asyncio
import logging
import warnings
from typing import TYPE_CHECKING, Callable

import aiohttp
from aiohttp import ClientSession, ClientTimeout

from pymexc.base_websocket import (
    FUTURES,
    SPOT,
    _WebSocketManager,
)

if TYPE_CHECKING:
    from pymexc._async.spot import HTTP

logger = logging.getLogger(__name__)


class _AsyncWebSocketManager(_WebSocketManager):
    endpoint: str

    def __init__(
        self,
        callback_function,
        ws_name,
        api_key=None,
        api_secret=None,
        subscribe_callback=None,
        ping_interval=20,
        ping_timeout=None,
        retries=10,
        restart_on_error=True,
        trace_logging=False,
        private_auth_expire=1,
        http_proxy_host=None,
        http_proxy_port=None,
        http_no_proxy=None,
        http_proxy_auth=None,
        http_proxy_timeout=None,
        loop=None,
        proto=True,
        extend_proto_body=False,
    ):
        super().__init__(
            callback_function,
            ws_name,
            api_key=api_key,
            api_secret=api_secret,
            subscribe_callback=subscribe_callback,
            ping_interval=ping_interval,
            ping_timeout=ping_timeout,
            retries=retries,
            restart_on_error=restart_on_error,
            trace_logging=trace_logging,
            private_auth_expire=private_auth_expire,
            http_proxy_host=http_proxy_host,
            http_proxy_port=http_proxy_port,
            http_no_proxy=http_no_proxy,
            http_proxy_auth=http_proxy_auth,
            http_proxy_timeout=http_proxy_timeout,
            proto=proto,
            extend_proto_body=extend_proto_body,
        )
        self.connected = False
        self.loop = loop or asyncio.get_event_loop()

        if ping_timeout:
            warnings.warn(
                "ping_timeout is deprecated for async websockets, please use just ping_interval.",
            )

    def exit(self):
        if getattr(self, "exited", False):
            return

        if hasattr(self, "ping_timer") and self.ping_timer:
            self.ping_timer.cancel()
            self.ping_timer = None

        async def _shutdown():
            if getattr(self, "ws", None) and not self.ws.closed:
                await self.ws.close()
            session = getattr(self, "session", None)
            if session and not session.closed:
                await session.close()

        try:
            running_loop = None
            try:
                running_loop = asyncio.get_running_loop()
            except RuntimeError:
                pass

            if self.loop and self.loop.is_running():
                if running_loop is self.loop:
                    self.loop.create_task(_shutdown())
                else:
                    asyncio.run_coroutine_threadsafe(_shutdown(), self.loop)
            else:
                asyncio.run(_shutdown())
        except Exception:
            logger.exception(f"WebSocket {self.ws_name} shutdown failed")
        finally:
            self.ws = None
            if hasattr(self, "session"):
                self.session = None
            self.exited = True

    async def _on_open(self):
        self.connected = True
        super()._on_open()

    async def _loop_recv(self):
        try:
            async for msg in self.ws:
                if msg.type in [aiohttp.WSMsgType.TEXT, aiohttp.WSMsgType.BINARY]:
                    await self._on_message(msg.data)
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    await self._on_error(msg)
                    break
        finally:
            await self._on_close()
            await self.session.close()

    async def _on_message(self, message: str | bytes):
        """
        Parse incoming messages.
        """
        _message = super()._on_message(message, parse_only=True)
        await self.callback(_message)

    def is_connected(self):
        return self.connected

    async def _connect(self, url):
        """
        Open websocket in a thread.
        """

        async def resubscribe_to_topics():
            if not self.subscriptions:
                # There are no subscriptions to resubscribe to, probably
                # because this is a brand new WSS initialisation so there was
                # no previous WSS connection.
                return

            for subscription_message in self.subscriptions.values():
                await self.ws.send_json(subscription_message)

        self.attempting_connection = True

        self.endpoint = url

        # Attempt to connect for X seconds.
        retries = self.retries
        if retries == 0:
            infinitely_reconnect = True
        else:
            infinitely_reconnect = False

        while (infinitely_reconnect or retries > 0) and not self.is_connected():
            logger.debug(f"WebSocket {self.ws_name} attempting connection... to {url}")

            self.session = ClientSession()
            timeout = ClientTimeout(total=60)
            self.ws = await self.session.ws_connect(
                url=url,
                proxy=f"http://{self.proxy_settings['http_proxy_host']}:{self.proxy_settings['http_proxy_port']}"
                if self.proxy_settings["http_proxy_host"]
                else None,
                proxy_auth=self.proxy_settings["http_proxy_auth"],
                timeout=timeout,
            )

            # parse incoming messages
            await self._on_open()
            self.loop.create_task(self._loop_recv())

            if not self.is_connected():
                # If connection was not successful, raise error.
                if not infinitely_reconnect and retries <= 0:
                    self.exit()
                    raise Exception(
                        f"WebSocket {self.ws_name} ({self.endpoint}) connection "
                        f"failed. Too many connection attempts. pymexc will no "
                        f"longer try to reconnect."
                    )

        logger.info(f"WebSocket {self.ws_name} connected")

        # If given an api_key, authenticate.
        if self.api_key and self.api_secret:
            await self._auth()

        await resubscribe_to_topics()
        self.ping_task = self.loop.create_task(self._ping_loop())

        self.attempting_connection = False

    async def _ping_loop(self):
        while True:
            try:
                await self.ws.send_str(self.custom_ping_message)
            except aiohttp.ClientPayloadError as e:
                await self._on_error(e)

            await asyncio.sleep(self.ping_interval)

    async def _auth(self):
        msg = super()._auth(parse_only=True)

        # Authenticate with API.
        await self.ws.send_json(msg)

    async def _on_error(self, error: Exception):
        super()._on_error(error, parse_only=True)

        # Reconnect.
        if self.restart_on_error and not self.attempting_connection:
            self._reset()
            await self._connect(self.endpoint)

    async def _on_close(self):
        self.connected = False
        super()._on_close()

    async def _process_normal_message(self, message: dict):
        callback_function, callback_data = super()._process_normal_message(message=message, parse_only=True)

        if callback_function is None:
            return

        await callback_function(callback_data)


# # # # # # # # # #
#                 #
#     FUTURES     #
#                 #
# # # # # # # # # #


class _FuturesWebSocketManager(_AsyncWebSocketManager):
    def __init__(self, ws_name, **kwargs):
        callback_function = (
            kwargs.pop("callback_function") if kwargs.get("callback_function") else self._handle_incoming_message
        )

        super().__init__(callback_function, ws_name, **kwargs)

    async def subscribe(self, topic, callback, params: dict = {}):
        subscription_args = {"method": topic, "param": params}
        callback_topic = self._topic(topic)
        self._check_callback_directory(callback_topic)

        while not self.is_connected():
            # Wait until the connection is open before subscribing.
            await asyncio.sleep(0.1)

        await self.ws.send_json(subscription_args)
        self.subscriptions[topic] = subscription_args
        self._set_callback(callback_topic, callback)
        self.last_subsctiption = callback_topic

    async def unsubscribe(self, method: str | Callable) -> None:
        if not method:
            return

        if isinstance(method, str):
            # remove callback
            self._pop_callback(method)
            # send unsub message
            await self.ws.send_json({"method": f"unsub.{method}", "param": {}})

            # remove subscription from list
            for topic, sub in self.subscriptions.items():
                if sub["method"] == f"sub.{method}":
                    self.subscriptions.pop(topic)
                    break

            logger.debug(f"Unsubscribed from {method}")
        else:
            # this is a func, get name
            topic_name = method.__name__.replace("_stream", "").replace("_", ".")

            return await self.unsubscribe(topic_name)

    async def _process_auth_message(self, message: dict):
        # If we get successful futures auth, notify user
        if message.get("data") == "success":
            logger.debug(f"Authorization for {self.ws_name} successful.")
            self.auth = True

        # If we get unsuccessful auth, notify user.
        elif message.get("data") != "success":  # !!!!
            logger.debug(f"Authorization for {self.ws_name} failed. Please check your API keys and restart.")

    async def _handle_incoming_message(self, message: dict):
        def is_auth_message():
            return message.get("channel", "") == "rs.login"

        def is_subscription_message():
            return message.get("channel", "").startswith("rs.sub") or message.get("channel", "") == "rs.personal.filter"

        def is_pong_message():
            return message.get("channel", "") in ("pong", "clientId")

        def is_error_message():
            return message.get("channel", "") == "rs.error"

        if is_auth_message():
            await self._process_auth_message(message)
        elif is_subscription_message():
            self._process_subscription_message(message)
        elif is_pong_message():
            pass
        elif is_error_message():
            logger.error(f"WebSocket return error: {message}")
        else:
            await self._process_normal_message(message)

    async def custom_topic_stream(self, topic, callback):
        return await self.subscribe(topic=topic, callback=callback)


class _FuturesWebSocket(_FuturesWebSocketManager):
    def __init__(
        self,
        api_key: str = None,
        api_secret: str = None,
        loop: asyncio.AbstractEventLoop = None,
        subscribe_callback: Callable = None,
        **kwargs,
    ):
        self.ws_name = "FuturesV1"
        self.endpoint = FUTURES
        loop = loop or asyncio.get_event_loop()

        if subscribe_callback:
            loop.create_task(self.connect())

        super().__init__(
            self.ws_name,
            api_key=api_key,
            api_secret=api_secret,
            loop=loop,
            subscribe_callback=subscribe_callback,
            **kwargs,
        )

    async def connect(self):
        if not self.is_connected():
            await self._connect(self.endpoint)

    async def _ws_subscribe(self, topic, callback, params: list = []):
        await self.connect()
        await self.subscribe(topic, callback, params)


# # # # # # # # # #
#                 #
#       SPOT      #
#                 #
# # # # # # # # # #


class _SpotWebSocketManager(_AsyncWebSocketManager):
    def __init__(self, ws_name, **kwargs):
        callback_function = (
            kwargs.pop("callback_function") if kwargs.get("callback_function") else self._handle_incoming_message
        )
        super().__init__(callback_function, ws_name, **kwargs)

        self.private_topics = ["account", "deals", "orders"]

    async def subscribe(self, topic: str, callback: Callable, params_list: list, interval: str = None):
        subscription_args = {
            "method": "SUBSCRIPTION",
            "params": [
                "@".join(
                    [f"spot@{topic}.v3.api" + (".pb" if self.proto else "")]
                    + ([str(interval)] if interval else [])
                    + list(map(str, params.values()))
                )
                for params in params_list
            ],
        }
        self._check_callback_directory(topic)

        while not self.is_connected():
            # Wait until the connection is open before subscribing.
            await asyncio.sleep(0.1)

        logger.debug(f"subscription_args: {subscription_args}")
        await self.ws.send_json(subscription_args)

        self.subscriptions[topic] = subscription_args
        print(self.subscriptions)

        self._set_callback(topic, callback)
        self.last_subsctiption = topic

    async def unsubscribe(self, *topics: str | Callable):
        if isinstance(topics, str):
            topics = [topics]

        elif isinstance(topics, tuple):
            topics = list(topics)

        for i in range(len(topics)):
            topic = topics[i]
            if isinstance(topic, str) and topic not in self.subscriptions:
                logger.error(f"[unsubscribe] Topic {topic} not found in subscriptions, skipping")
                topics.pop(i)
                continue

        if all([isinstance(topic, str) for topic in topics]):
            # send unsub message
            unsubscribe_message = {
                "method": "UNSUBSCRIPTION",
                "params": [],
            }

            # remove callbacks
            # remove subscriptions from list
            for topic in topics:
                self._pop_callback(topic)
                unsubscribe_message["params"] += self.subscriptions[topic]["params"]
                self.subscriptions.pop(topic)

            logger.debug(f"Unsubscribe message: {unsubscribe_message}")
            await self.ws.send_json(unsubscribe_message)

            logger.info(f"Unsubscribed from {topics}")
        else:
            # some funcs in list
            topics = [
                x.__name__.replace("_stream", "").replace("_", ".") if getattr(x, "__name__", None) else x
                #
                for x in topics
            ]

            topics = [self.func_to_topic[x] if x in self.func_to_topic else x for x in topics]
            return await self.unsubscribe(*topics)

    async def _handle_incoming_message(self, message):
        logger.debug(f"_handle_incoming_message: {message}")

        def is_subscription_message():
            if message.get("id") == 0 and message.get("code") == 0:
                return True
            else:
                return False

        if isinstance(message, dict) and is_subscription_message():
            self._process_subscription_message(message)
        else:
            await self._process_normal_message(message)

    async def custom_topic_stream(self, topic, callback):
        return await self.subscribe(topic=topic, callback=callback)


class _SpotWebSocket(_SpotWebSocketManager):
    listenKey: str
    http: "HTTP"

    def __init__(
        self,
        endpoint: str = SPOT,
        api_key: str = None,
        api_secret: str = None,
        loop: asyncio.AbstractEventLoop = None,
        **kwargs,
    ):
        self.ws_name = "SpotV3"
        self.endpoint = endpoint
        loop = loop or asyncio.get_event_loop()

        super().__init__(self.ws_name, api_key=api_key, api_secret=api_secret, loop=loop, **kwargs)

    async def _ws_subscribe(self, topic, callback, params: list = [], interval: str = None):
        if isinstance(topic, str) and topic.startswith("private."):
            ensure_listen_key = getattr(self, "_ensure_listen_key", None)
            logger.debug(f"ensure_listen_key: {ensure_listen_key}")
            if ensure_listen_key:
                await ensure_listen_key()

        if not self.is_connected():
            await self._connect(self.endpoint)

        logger.info(f"Subscribing to topic: {topic} | params: {params} | interval: {interval}")
        await self.subscribe(topic, callback, params, interval)



# pytest . -v -s

import logging
import threading
import time

import pytest
from dotenv import dotenv_values

from pymexc.proto import (
    PrivateAccountV3Api,
    PrivateDealsV3Api,
    PrivateOrdersV3Api,
    PublicAggreBookTickerV3Api,
    PublicAggreDealsV3Api,
    PublicAggreDepthsV3Api,
    PublicBookTickerBatchV3Api,
    PublicLimitDepthsV3Api,
    PublicMiniTickerV3Api,
    PublicMiniTickersV3Api,
    PublicSpotKlineV3Api,
)
from pymexc.spot import HTTP, WebSocket

env = dotenv_values(".env")
api_key = env.get("API_KEY")
api_secret = env.get("API_SECRET")

if not api_key or not api_secret:
    pytest.skip("API credentials are required to run WebSocket tests", allow_module_level=True)

PRINT_RESPONSE = True
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s:%(lineno)d - %(levelname)s - %(message)s",
)
logger = logging.getLogger(__name__)


@pytest.fixture
def ws_client():
    client = WebSocket(api_key=api_key, api_secret=api_secret, proto=True)
    logger.info("WebSocket client created: %s", client)
    yield client
    try:
        client.exit()
    except Exception as exc:
        logger.warning("Failed to close WebSocket client: %s", exc)
    time.sleep(1)


def make_mxusdt_order(time_wait: int = 2):
    client = HTTP(api_key=api_key, api_secret=api_secret)
    time.sleep(time_wait)
    order_response = client.order("MXUSDT", "BUY", "LIMIT", quantity=20, price=0.1)
    logger.info("MXUSDT Order Response: %s", order_response)
    time.sleep(time_wait)
    cancel_order_response = client.cancel_order("MXUSDT", order_response["orderId"])
    logger.info("MXUSDT Cancel Order Response: %s", cancel_order_response)
    time.sleep(time_wait)


def _wait_or_fail(event: threading.Event, timeout: float, message: str):
    if not event.wait(timeout):
        pytest.fail(message)


def test_deals_stream(ws_client: WebSocket):
    messages: list[PublicAggreDealsV3Api] = []
    done = threading.Event()

    def handle_message(msg: PublicAggreDealsV3Api):
        messages.append(msg)
        logger.info("Received deals message: %s", msg)
        if len(messages) >= 2:
            ws_client.unsubscribe(ws_client.deals_stream)
            done.set()

    ws_client.deals_stream(handle_message, "BTCUSDT", interval="100ms")

    _wait_or_fail(done, 10, "Не получили достаточно сообщений сделки за отведённое время")
    assert len(messages) >= 2
    for msg in messages:
        assert isinstance(msg, PublicAggreDealsV3Api)
        assert msg.deals
        assert float(msg.deals[0].price) > 0
        assert float(msg.deals[0].quantity) > 0
        assert msg.deals[0].tradeType > 0
        assert int(msg.deals[0].time) > 0


def test_kline_stream(ws_client: WebSocket):
    messages: list[PublicSpotKlineV3Api] = []
    done = threading.Event()

    def handle_message(msg: PublicSpotKlineV3Api):
        messages.append(msg)
        logger.info("Received kline message: %s", msg)
        if len(messages) >= 2:
            ws_client.unsubscribe(ws_client.kline_stream)
            done.set()

    ws_client.kline_stream(handle_message, "BTCUSDT", interval="Min1")

    _wait_or_fail(done, 10, "Не получили достаточно сообщений клина за отведённое время")
    assert len(messages) >= 2
    for msg in messages:
        assert isinstance(msg, PublicSpotKlineV3Api)
        assert msg.interval == "Min1"
        assert int(msg.windowStart) > 0
        assert float(msg.openingPrice) > 0
        assert float(msg.closingPrice) > 0
        assert float(msg.highestPrice) > 0
        assert float(msg.lowestPrice) > 0
        assert float(msg.volume) > 0
        assert float(msg.amount) > 0
        assert int(msg.windowEnd) > 0


def test_increase_depth_stream(ws_client: WebSocket):
    messages: list[PublicAggreDepthsV3Api] = []
    done = threading.Event()

    def handle_message(msg: PublicAggreDepthsV3Api):
        messages.append(msg)
        if len(messages) >= 2:
            ws_client.unsubscribe(ws_client.depth_stream)
            done.set()

    ws_client.depth_stream(handle_message, "BTCUSDT")

    _wait_or_fail(done, 10, "Не получили достаточно сообщений об агрегированной глубине")
    assert len(messages) >= 2
    for msg in messages:
        assert isinstance(msg, PublicAggreDepthsV3Api)
        assert msg.eventType == "spot@public.aggre.depth.v3.api.pb@100ms"
        assert msg.fromVersion is not None
        assert msg.asks
        assert msg.bids


def test_limit_depth_stream(ws_client: WebSocket):
    messages: list[PublicLimitDepthsV3Api] = []
    done = threading.Event()

    def handle_message(msg: PublicLimitDepthsV3Api):
        messages.append(msg)
        logger.info("Received limit depth message: %s", msg)
        if len(messages) >= 2:
            ws_client.unsubscribe(ws_client.limit_depth_stream)
            done.set()

    ws_client.limit_depth_stream(handle_message, "BTCUSDT", 5)

    _wait_or_fail(done, 10, "Не получили достаточно сообщений о лимитной глубине")
    assert len(messages) >= 2
    for msg in messages:
        assert isinstance(msg, PublicLimitDepthsV3Api)
        assert msg.eventType == "spot@public.limit.depth.v3.api.pb"
        assert msg.version is not None
        assert msg.asks
        assert msg.bids


def test_book_ticker_stream(ws_client: WebSocket):
    messages: list[PublicAggreBookTickerV3Api] = []
    done = threading.Event()

    def handle_message(msg: PublicAggreBookTickerV3Api):
        messages.append(msg)
        if len(messages) >= 2:
            ws_client.unsubscribe(ws_client.book_ticker_stream)
            done.set()

    ws_client.book_ticker_stream(handle_message, "BTCUSDT")

    _wait_or_fail(done, 10, "Не получили достаточно сообщений book ticker")
    assert len(messages) >= 2
    for msg in messages:
        assert isinstance(msg, PublicAggreBookTickerV3Api)
        assert msg.bidPrice is not None and float(msg.bidPrice) > 0
        assert msg.bidQuantity is not None and float(msg.bidQuantity) > 0
        assert msg.askPrice is not None and float(msg.askPrice) > 0
        assert msg.askQuantity is not None and float(msg.askQuantity) > 0


def test_book_ticker_batch_stream(ws_client: WebSocket):
    messages: list[PublicBookTickerBatchV3Api] = []
    done = threading.Event()

    def handle_message(msg: PublicBookTickerBatchV3Api):
        messages.append(msg)
        if len(messages) >= 2:
            ws_client.unsubscribe(ws_client.book_ticker_batch_stream)
            done.set()

    ws_client.book_ticker_batch_stream(handle_message, ["BTCUSDT"])

    _wait_or_fail(done, 10, "Не получили достаточно сообщений batch book ticker")
    assert len(messages) >= 2
    for msg in messages:
        assert isinstance(msg, PublicBookTickerBatchV3Api)
        assert msg.items
        first = msg.items[0]
        assert first.bidPrice is not None and float(first.bidPrice) > 0
        assert first.bidQuantity is not None and float(first.bidQuantity) > 0
        assert first.askPrice is not None and float(first.askPrice) > 0
        assert first.askQuantity is not None and float(first.askQuantity) > 0


def test_mini_tickers_stream(ws_client: WebSocket):
    messages: list[PublicMiniTickersV3Api] = []
    done = threading.Event()

    def handle_message(msg: PublicMiniTickersV3Api):
        messages.append(msg)
        logger.info(f"Received mini tickers message: {msg}")
        if len(messages) >= 1:  # Mini tickers push every 3 seconds, so wait for at least 1
            ws_client.unsubscribe(ws_client.mini_tickers_stream)
            done.set()

    ws_client.mini_tickers_stream(handle_message, timezone="UTC+8")

    _wait_or_fail(done, 15, "Did not receive mini tickers messages")  # Increased timeout for 3-second interval
    assert len(messages) >= 1
    for msg in messages:
        assert isinstance(msg, PublicMiniTickersV3Api)
        assert hasattr(msg, "items") and msg.items
        assert len(msg.items) > 0
        first = msg.items[0]
        assert hasattr(first, "symbol") and first.symbol
        assert hasattr(first, "price") and first.price is not None


def test_mini_ticker_stream(ws_client: WebSocket):
    messages: list[PublicMiniTickerV3Api] = []
    done = threading.Event()

    def handle_message(msg: PublicMiniTickerV3Api):
        messages.append(msg)
        logger.info(f"Received mini ticker message: {msg}")
        if len(messages) >= 1:  # Mini ticker pushes every 3 seconds, so wait for at least 1
            ws_client.unsubscribe(ws_client.mini_ticker_stream)
            done.set()

    ws_client.mini_ticker_stream(handle_message, "BTCUSDT", timezone="UTC+8")

    _wait_or_fail(done, 15, "Did not receive mini ticker messages")  # Increased timeout for 3-second interval
    assert len(messages) >= 1
    for msg in messages:
        assert isinstance(msg, PublicMiniTickerV3Api)
        assert hasattr(msg, "symbol") and msg.symbol == "BTCUSDT"
        assert hasattr(msg, "price") and msg.price is not None


def test_account_update(ws_client: WebSocket):
    messages: list[PrivateAccountV3Api] = []
    done = threading.Event()

    def handle_message(msg: PrivateAccountV3Api):
        messages.append(msg)
        logger.info("Received account update: %s", msg)
        ws_client.unsubscribe(ws_client.account_update)
        done.set()

    ws_client.account_update(handle_message)

    order_thread = threading.Thread(target=make_mxusdt_order, kwargs={"time_wait": 2})
    order_thread.start()

    _wait_or_fail(done, 20, "Не получили сообщение об обновлении аккаунта")
    order_thread.join(timeout=10)

    assert messages
    for msg in messages:
        assert isinstance(msg, PrivateAccountV3Api)
        assert msg.vcoinName == "USDT"
        assert msg.coinId
        assert msg.balanceAmount is not None
        assert msg.balanceAmountChange is not None
        assert msg.frozenAmount is not None
        assert msg.frozenAmountChange is not None
        assert msg.type
        assert msg.time > 0


def test_account_deals(ws_client: WebSocket):
    messages: list[PrivateDealsV3Api] = []
    done = threading.Event()

    def handle_message(msg: PrivateDealsV3Api):
        messages.append(msg)
        logger.info("Received account deal: %s", msg)
        if len(messages) >= 2:
            ws_client.unsubscribe(ws_client.account_deals)
            done.set()

    ws_client.account_deals(handle_message)

    order_thread = threading.Thread(target=make_mxusdt_order, kwargs={"time_wait": 2})
    order_thread.start()

    # Ждём, но не валимся по таймауту — биржа может не выдать сделки
    done.wait(20)
    ws_client.unsubscribe(ws_client.account_deals)
    order_thread.join(timeout=10)
    time.sleep(2)

    if PRINT_RESPONSE:
        logger.info("Account deals messages: %s", messages)

    assert len(messages) >= 0
    for msg in messages:
        assert isinstance(msg, PrivateDealsV3Api)
        assert float(msg.price) > 0
        assert float(msg.quantity) > 0
        assert msg.tradeType > 0
        assert int(msg.time) > 0


def test_account_orders(ws_client: WebSocket):
    messages: list[PrivateOrdersV3Api] = []
    done = threading.Event()

    def handle_message(msg: PrivateOrdersV3Api):
        messages.append(msg)
        logger.info("Received account order: %s", msg)
        ws_client.unsubscribe(ws_client.account_orders)
        done.set()

    ws_client.account_orders(handle_message)

    order_thread = threading.Thread(target=make_mxusdt_order, kwargs={"time_wait": 2})
    order_thread.start()

    _wait_or_fail(done, 20, "Не получили сообщение по заявкам аккаунта")
    order_thread.join(timeout=10)

    assert messages
    for msg in messages:
        assert isinstance(msg, PrivateOrdersV3Api)
        assert msg.id is not None
        assert msg.clientId is not None
        assert msg.price is not None
        assert msg.quantity is not None
        assert msg.amount is not None
        assert msg.avgPrice is not None
        assert msg.orderType is not None
        assert msg.tradeType is not None
        assert msg.isMaker is not None
        assert msg.remainAmount is not None
        assert msg.remainQuantity is not None
        assert msg.lastDealQuantity is not None
        assert msg.cumulativeQuantity is not None
        assert msg.cumulativeAmount is not None
        assert msg.status is not None
        assert msg.createTime is not None
        assert msg.market is not None
        assert msg.triggerType is not None
        assert msg.triggerPrice is not None
        assert msg.state is not None
        assert msg.ocoId is not None
        assert msg.routeFactor is not None
        assert msg.symbolId is not None
        assert msg.marketId is not None
        assert msg.marketCurrencyId is not None
        assert msg.currencyId is not None


def test_multiple_subscriptions(ws_client: WebSocket):
    messages: list = []
    done = threading.Event()

    def handle_message(msg):
        messages.append(msg)
        logger.info(f"Received message type: {type(msg)}")
        if len(messages) >= 4:
            ws_client.unsubscribe(ws_client.deals_stream, ws_client.book_ticker_stream)
            done.set()

    ws_client.deals_stream(handle_message, "BTCUSDT")
    ws_client.book_ticker_stream(handle_message, "BTCUSDT")

    _wait_or_fail(done, 15, "Не получили сообщения от обоих потоков")
    assert len(messages) >= 4

    message_types = {type(msg) for msg in messages}
    assert PublicAggreDealsV3Api in message_types
    assert PublicAggreBookTickerV3Api in message_types

    for msg in messages:
        if isinstance(msg, PublicAggreDealsV3Api):
            assert msg.deals
            assert msg.deals[0].price is not None and float(msg.deals[0].price) > 0
            assert msg.deals[0].quantity is not None and float(msg.deals[0].quantity) > 0
            assert msg.deals[0].tradeType is not None and msg.deals[0].tradeType > 0
            assert msg.deals[0].time is not None and int(msg.deals[0].time) > 0

        if isinstance(msg, PublicAggreBookTickerV3Api):
            assert msg.bidPrice is not None and float(msg.bidPrice) > 0
            assert msg.bidQuantity is not None and float(msg.bidQuantity) > 0
            assert msg.askPrice is not None and float(msg.askPrice) > 0
            assert msg.askQuantity is not None and float(msg.askQuantity) > 0
